module envision.canvas;

private
{
    import std.stream;
}

/**
* A colour.  Currently only stored in 32bpp ARGB format.
* TODO: rectify this, so that pixel formats can vary any which way.
*/
struct Colour
{
    align(1) ubyte b, g, r, a;

    static
    {
        /** Named constructor idiom, since D doesn't allow structs to have
        *  constructors.
        * @param R Red intensity.
        * @param G Green intensity.
        * @param B Blue intensity.
        * @param A Alpha level. (255 is opaque, 0 is completely transparent)
        * @return The colour.
        */
        Colour create(ubyte R, ubyte G, ubyte B, ubyte A)
        {
            Colour c;
            c.a = A;
            c.r = R;
            c.g = G;
            c.b = B;
            return c;
        }

        /** Same as above, but sets alpha to the maximum (the colour is opaque).
        * @param R Red intensity.
        * @param G Green intensity.
        * @param B Blue intensity.
        * @return The colour.
        */
        Colour create(ubyte R, ubyte G, ubyte B)
        {
            Colour c;
            c.a = 255;
            c.r = R;
            c.g = G;
            c.b = B;
            return c;
        }

        Colour BLACK = { b:  0, g:  0, r:  0, a:255 };
        Colour WHITE = { b:255, g:255, r:255, a:255 };
    }
}

/**
 * Software representation of an image.
 */
final class Canvas
{
private:
    Colour[] _pixels;
    uint _width;
    uint _height;

public:
    this(uint w, uint h)
    in  {   assert(w > 0 && h > 0); }
    body
    {
        _width = w;
        _height = h;
        _pixels.length = w * h;
        _pixels[] = Colour.WHITE;
    }

    this(Colour* p, uint w, uint h)
    in  {   assert(p != null && w > 0 && h > 0);    }
    body
    {
        _pixels.length = w * h;
        _pixels = p[0 .. (w * h)];

        _width = w;
        _height = h;
    }

    /** @return A pointer to the raw pixel data.  Do not delete this pointer! */
    Colour* pixels()
    {
        return &_pixels[0];
    }

    /** @return The width of the image. */
    uint width()
    {
        return _width;
    }

    /** @return The height of the image. */
    uint height()
    {
        return _height;
    }

    /** @return Returns the pixel at (x, y) */
    Colour getPixel(int x, int y)
    {
        if (x < 0 || y < 0 || x >= _width || y >= _height)
            return Colour.create(0, 0, 0, 0);
        else
            return _pixels[y * _width + x];
    }

    /**
    * Returns a pointer to the pixel at (x,y)
    * Used by things that need fast sequential access to a bunch of
    * pixels.  Pixels are guaranteed to be adjacent within a column,
    * but rows of pixels are not. (the pitch of the Canvas may not
    * actually be equal to the width of the image)
    */
    Colour* getPointer(uint x, uint y)
    in  { assert(x < _width && y < _height); }
    body
    {
        return &_pixels[y * _width + x];
    }

    /**
    * @return The number of bytes required to store one pixel.
    */
    int pixelSize()
    {
        return 4;
    }
}
