/**
 * libPNG interface.
 *
 * Adapted version of Burton Raydon's image loader for dig.
 * Loads a PNG image into a software buffer, where it can be manipulated
 * freely.
 *
 * Tweaks by Andy Friesen.
 *
 * Some pluggable factory things could be done here so that more image
 * formats could be added more easily.
 *
 * I'd also like to eventually support more pixel formats than 32bpp ARGB.
 * I was thinking that I should store the pixel data as a void*, then have
 * a PixelFormatStrategy, which would be responsible for returning pixels
 * in some requested format, without ever forcing the user to manually
 * deal with pixel format conversions.
 *
 * Distribution / Code rights:
 *  Use this source code in any fashion you see fit.  Giving me credit where
 *  credit is due is optional, depending on your own levels of integrity and
 *  honesty.
 *
 */

module envision.libpng;

private
{
    import std.c.windows.windows;
    import std.c.stdio;
    import std.stream;
}

private import envision.canvas;

private
{
    HANDLE library;
    const char[] libraryName = "libpng.dll";

    static this()
    {
        library = LoadLibraryA(libraryName);
        if (library == (HANDLE) 0)
            throw new Exception("Unable to load " ~ libraryName);

        load(cast(void**)&png_create_info_struct, "png_create_info_struct");
        load(cast(void**)&png_create_read_struct, "png_create_read_struct");
        load(cast(void**)&png_create_write_struct, "png_create_write_struct");
        load(cast(void**)&png_destroy_read_struct, "png_destroy_read_struct");
        load(cast(void**)&png_destroy_write_struct, "png_destroy_write_struct");
        load(cast(void**)&png_get_IHDR, "png_get_IHDR");
        load(cast(void**)&png_get_PLTE, "png_get_PLTE");
        load(cast(void**)&png_get_io_ptr, "png_get_io_ptr");
        load(cast(void**)&png_get_text, "png_get_text");
        load(cast(void**)&png_get_rowbytes, "png_get_rowbytes");
        load(cast(void**)&png_init_io, "png_init_io");
        load(cast(void**)&png_read_image, "png_read_image");
        load(cast(void**)&png_read_info, "png_read_info");
        load(cast(void**)&png_read_row, "png_read_row");
        load(cast(void**)&png_set_IHDR, "png_set_IHDR");
        load(cast(void**)&png_set_packing, "png_set_packing");
        load(cast(void**)&png_set_read_fn, "png_set_read_fn");
        load(cast(void**)&png_set_sig_bytes, "png_set_sig_bytes");
        load(cast(void**)&png_set_strip_16, "png_set_strip_16");
        load(cast(void**)&png_set_text, "png_set_text");
        load(cast(void**)&png_set_write_fn, "png_set_write_fn");
        load(cast(void**)&png_sig_cmp, "png_sig_cmp");
        load(cast(void**)&png_write_end, "png_write_end");
        load(cast(void**)&png_write_image, "png_write_image");
        load(cast(void**)&png_write_info, "png_write_info");
        load(cast(void**)&png_malloc, "png_malloc");
        load(cast(void**)&png_set_rows, "png_set_rows");
        load(cast(void**)&png_write_png, "png_write_png");
    }

    // Wish I didn't have to use void**
    // inout void* would be so much nicer. :(
    void load (void** pointer, char[] name)
    {
        *pointer = GetProcAddress (library, toStringz (name));
        if (*pointer === null)
            throw new Exception("couldn't find " ~ name ~ " in " ~ libraryName);
    }

    // Pixel format conversion routines.
    // length is the number of pixels, not bytes
    Colour[] i8ToRGBA(ubyte[] src, int length, Colour[] palette)
    {
        Colour[] dest = new Colour[length];
        for (uint i = 0; i < length; i++)
        {
            ubyte p = src[i];
            dest[i] = palette[p];
        }
        return dest;
    }

    Colour[] RGBtoRGBA(ubyte[] src, int length)
    {
        ubyte* p = &src[0];
        Colour[] dest = new Colour[length];
        for (uint i = 0; i < length; i++)
        {
            ubyte b = *p++;
            ubyte g = *p++;
            ubyte r = *p++;
            dest[i] = Colour.create(r, g, b);
        }
        return dest;
    }

    Colour[] grayScaleToRGBA(ubyte[] src, int length)
    {
        Colour[] dest = new Colour[length];
        for (int i = 0; i < length; i++)
            dest[i] = Colour.create(src[i], src[i], src[i]);
        return dest;
    }

    Colour[] grayScaleAlphaToRGBA(ubyte[] src, int length)
    {
        Colour[] dest = new Colour[length];
        int j = 0;
        for (uint i = 0; i < length; i++)
        {
            ubyte l = src[j++];
            ubyte a = src[j++];
            dest[i] = Colour.create(l, l, l, a);
        }
        return dest;
    }

    // heh.  Maybe I should just cast or something.
    Colour[] RGBAtoRGBA(inout ubyte[] src, int length)
    {
        ubyte* p = &src[0];
        Colour[] dest = new Colour[length];
        for (uint i = 0; i < length; i++)
        {
            ubyte r = *p++;
            ubyte g = *p++;
            ubyte b = *p++;
            ubyte a = *p++;
            dest[i] = Colour.create(r, g, b, a);
        }

        return dest;
    }
}

private extern (C)
{
    struct png_struct
    {
    }

    struct png_info
    {
        uint width;
        uint height;
        uint valid;
        uint rowbytes;
        void* palette;
        ushort num_palette;
        ushort num_trans;
        ubyte bit_depth;;
        ubyte color_type;
        ubyte compression_type;
        ubyte filter_type;
        ubyte interlace_type;
        ubyte channels;
        ubyte pixel_depth;
        ubyte spare_byte;
        ubyte[8] signature;

        // Other stuff here too.
    }

    struct png_text
    {
        int compression; /* compression value:
                           -1: tEXt, none
                            0: zTXt, deflate
                            1: iTXt, none
                            2: iTXt, deflate  */
        char *key; /* keyword, 1-79 character description of "text" */
        char *text; /* comment, may be an empty string (ie "") */
        int text_length; /* length of the text string */
        int itxt_length; /* length of the itxt string */
        char *lang; /* language code, 1-79 characters */
        char *lang_key; /* keyword translated UTF-8 string, 0 or more chars */
    }

    struct png_color
    {
        ubyte red;
        ubyte green;
        ubyte blue;
    }

    typedef void (*png_error) (png_struct *, char *);
    typedef void (*png_flush) (png_struct *);
    typedef void (*png_rw) (png_struct *, byte *, int);

    const char *PNG_LIBPNG_VER_STRING = "1.0.6";

    /* These describe the color_type field in png_info. */
    /* color type masks */
    const int PNG_COLOR_MASK_PALETTE = 1;
    const int PNG_COLOR_MASK_COLOR = 2;
    const int PNG_COLOR_MASK_ALPHA = 4;

    /* color types.  Note that not all combinations are legal */
    const int PNG_COLOR_TYPE_GRAY = 0;
    const int PNG_COLOR_TYPE_PALETTE = (PNG_COLOR_MASK_COLOR | PNG_COLOR_MASK_PALETTE);
    const int PNG_COLOR_TYPE_RGB = (PNG_COLOR_MASK_COLOR);
    const int PNG_COLOR_TYPE_RGB_ALPHA = (PNG_COLOR_MASK_COLOR | PNG_COLOR_MASK_ALPHA);
    const int PNG_COLOR_TYPE_GRAY_ALPHA = (PNG_COLOR_MASK_ALPHA);

    /* These are for the interlacing type.  These values should NOT be changed. */
    const int PNG_INTERLACE_NONE = 0; /* Non-interlaced image */
    const int PNG_INTERLACE_ADAM7 = 1; /* Adam7 interlacing */
    const int PNG_INTERLACE_LAST = 2; /* Not a valid value */

    /* This is for compression type. PNG 1.0-1.2 only define the single type. */
    const int PNG_COMPRESSION_TYPE_BASE = 0; /* Deflate method 8, 32K window */
    const int PNG_COMPRESSION_TYPE_DEFAULT = PNG_COMPRESSION_TYPE_BASE;

    /* This is for filter type. PNG 1.0-1.2 only define the single type. */
    const int PNG_FILTER_TYPE_BASE = 0; /* Single row per-byte filtering */
    const int PNG_FILTER_TYPE_DEFAULT = PNG_FILTER_TYPE_BASE;

    /* Supported compression types for text in PNG files (tEXt, and zTXt).
     * The values of the PNG_TEXT_COMPRESSION_ defines should NOT be changed. */
    const int PNG_TEXT_COMPRESSION_NONE_WR = -3;
    const int PNG_TEXT_COMPRESSION_zTXt_WR = -2;
    const int PNG_TEXT_COMPRESSION_NONE = -1;
    const int PNG_TEXT_COMPRESSION_zTXt = 0;
    const int PNG_ITXT_COMPRESSION_NONE = 1;
    const int PNG_ITXT_COMPRESSION_zTXt = 2;
    const int PNG_TEXT_COMPRESSION_LAST = 3;  /* Not a valid value */

    const int PNG_INFO_gAMA = 0x0001;
    const int PNG_INFO_sBIT = 0x0002;
    const int PNG_INFO_cHRM = 0x0004;
    const int PNG_INFO_PLTE = 0x0008;
    const int PNG_INFO_tRNS = 0x0010;
    const int PNG_INFO_bKGD = 0x0020;
    const int PNG_INFO_hIST = 0x0040;
    const int PNG_INFO_pHYs = 0x0080;
    const int PNG_INFO_oFFs = 0x0100;
    const int PNG_INFO_tIME = 0x0200;
    const int PNG_INFO_pCAL = 0x0400;
    const int PNG_INFO_sRGB = 0x0800;   // GR-P, 0.96a
    const int PNG_INFO_iCCP = 0x1000;   // ESR, 1.0.6
    const int PNG_INFO_sPLT = 0x2000;   // ESR, 1.0.6
    const int PNG_INFO_sCAL = 0x4000;   // ESR, 1.0.6
    const int PNG_INFO_IDAT = 0x8000L;  // ESR, 1.0.6

    const int PNG_TRANSFORM_IDENTITY       = 0x0000;   // read and write
    const int PNG_TRANSFORM_STRIP_16       = 0x0001;   // read only
    const int PNG_TRANSFORM_STRIP_ALPHA    = 0x0002;   // read only
    const int PNG_TRANSFORM_PACKING        = 0x0004;   // read and write
    const int PNG_TRANSFORM_PACKSWAP       = 0x0008;   // read and write
    const int PNG_TRANSFORM_EXPAND         = 0x0010;   // read only
    const int PNG_TRANSFORM_INVERT_MONO    = 0x0020;   // read and write
    const int PNG_TRANSFORM_SHIFT          = 0x0040;   // read and write
    const int PNG_TRANSFORM_BGR            = 0x0080;   // read and write
    const int PNG_TRANSFORM_SWAP_ALPHA     = 0x0100;   // read and write
    const int PNG_TRANSFORM_SWAP_ENDIAN    = 0x0200;   // read and write
    const int PNG_TRANSFORM_INVERT_ALPHA   = 0x0400;   // read and write
    const int PNG_TRANSFORM_STRIP_FILLER   = 0x0800;   // WRITE only



    png_info* (*png_create_info_struct) (png_struct* png);
    png_struct* (*png_create_read_struct) (char* user_png_ver, void* error_ptr, png_error* error_fn, png_error* warn_fn);
    png_struct* (*png_create_write_struct) (char* user_png_ver, void* error_ptr, png_error* error_fn, png_error* warn_fn);
    void  (*png_destroy_read_struct) (png_struct** png, png_info** info, png_info** end);
    void  (*png_destroy_write_struct) (png_struct** png, png_info** info);
    uint  (*png_get_IHDR) (png_struct* png, png_info* info, uint* width, uint* height, int* bit_depth, int* color_type, int* interlace_type, int* compression_type, int* filter_type);
    uint  (*png_get_PLTE) (png_struct* png, png_info* info, png_color** palette, int* num_palette);
    void* (*png_get_io_ptr) (png_struct* png);
    uint  (*png_get_text) (png_struct* png, png_info* info, png_text** text, int* num_text);
    uint  (*png_get_rowbytes) (png_struct* png, png_info* info);
    void  (*png_init_io) (png_struct* png, FILE* fp);
    void  (*png_read_image) (png_struct* png, byte** image);
    void  (*png_read_info) (png_struct* png, png_info* info);
    void  (*png_read_row) (png_struct* png, ubyte* row, ubyte* display_row);
    void  (*png_set_IHDR) (png_struct* png, png_info* info, uint width, uint height, int bit_depth, int color_type, int interlace_type, int compression_type, int filter_type);
    void  (*png_set_packing) (png_struct* png);
    void  (*png_set_read_fn) (png_struct* png, void* io_ptr, png_rw read_data_fn);
    void  (*png_set_sig_bytes) (png_struct* png, int num_bytes);
    void  (*png_set_strip_16) (png_struct* png);
    void  (*png_set_text) (png_struct* png, png_info* info, png_text* text, int num_text);
    void  (*png_set_write_fn) (png_struct* png, void* io_ptr, png_rw write_data_fn, png_flush output_flush_fn);
    int   (*png_sig_cmp) (ubyte* sig, int start, int num_to_check);
    void  (*png_write_end) (png_struct* png, png_info* info);
    void  (*png_write_image) (png_struct* png, byte** image);
    void  (*png_write_info) (png_struct* png, png_info* info);
    void* (*png_malloc)(png_struct* png, uint size);
    void  (*png_set_rows)(png_struct* png, png_info* info, void** rows);
    void  (*png_write_png)(png_struct* png, png_info* info, int transforms, void* params);

    void readPNG(png_struct* png, byte* data, int length)
    {
        Stream stream;

        stream = cast (Stream) png_get_io_ptr (png);
        stream.readExact (data, length);
    }

    void writePNG(png_struct* png, byte* data, int length)
    {
        Stream stream = cast(Stream)png_get_io_ptr(png);
        stream.writeExact(data, length);
    }

    void flushPNG(png_struct* png)
    {}
}

/** Loads an image.
 * @param fileName Filename of the image to load.
 * @return The loaded image.
 * @throw Exception if an error occurs.
 */
Canvas loadPNG(char[] fileName)
{
    Stream stream = new File(fileName, FileMode.In);
    Canvas img = loadPNG(stream);
    delete stream;

    return img;
}

/** Saves an image.
 * @param fileName Filename to save the image to.
 */
void savePNG(char[] fileName, Canvas canvas)
{
    File stream = new File();
    stream.create(fileName);
    savePNG(stream, canvas);
    delete stream;
}

/** Loads an image.
 * @param stream The stream from which to load the image.
 * @return The image.
 * @throw Exception if the image cannot be loaded.
 */
Canvas loadPNG(Stream stream)
{
    ubyte[8] header;
    if (stream.read(header) != 8)
        throw new Exception("Unable to read PNG header.");

    if (png_sig_cmp(header, 0, header.length) != 0)
        throw new Exception("File is not a PNG image.");

    png_struct* png = null;
    png_info* info = null;
    png_info* end = null;

    try
    {
        // Initial libpng bootup.
        png = png_create_read_struct (PNG_LIBPNG_VER_STRING, null, null, null);
        if (png === null)
            throw new Exception("Couldn't allocate png struct.");

        info = png_create_info_struct (png);
        if (info === null)
            throw new Exception("Couldn't allocate PNG info block");

        end = png_create_info_struct (png);
        if (end === null)
            throw new Exception("Couldn't allocate PNG end info block");

        png_set_read_fn(png, cast(void*) stream, &readPNG);
        //png_init_io (png, file);
        png_set_sig_bytes(png, header.length);

        uint width, height;
        int depth;
        int colorType;
        int filterType;
        int compressionType;
        int interlaceType;

        png_read_info(png, info);
        png_get_IHDR(png, info, &width, &height, &depth, &colorType,
            &interlaceType, &compressionType, &filterType);

        int pitch = png_get_rowbytes(png, info);
        ubyte[] pixels = new ubyte[width * height * depth];
        Colour[] palette = null;

        // extract and convert palette
        if (colorType == PNG_COLOR_TYPE_PALETTE)
        {
            png_color* colours;
            int count;
            png_get_PLTE(png, info, &colours, &count);
            palette = new Colour[count];

            for (int c = 0; c < count; c++)
            {
                png_color* s = &colours[c];
                palette[c] = Colour.create(s.blue, s.green, s.red); // Why?!  Should be red green blue. @_@
            }
        }

        // read the actual pixels
        {
            ubyte* row = &pixels[0];
            for (int y = 0; y < height; y++)
            {
                png_read_row(png, row, null);
                row += pitch;
            }
        }

        Colour[] rgba;
        switch (colorType)
        {
            case PNG_COLOR_TYPE_GRAY:           rgba = grayScaleToRGBA(pixels, width * height);         break;
            case PNG_COLOR_TYPE_GRAY_ALPHA:     rgba = grayScaleAlphaToRGBA(pixels, width * height);    break;
            case PNG_COLOR_TYPE_PALETTE:        rgba = i8ToRGBA(pixels, width * height, palette);       break;
            case PNG_COLOR_TYPE_RGB:            rgba = RGBtoRGBA(pixels, width * height);               break;
            case PNG_COLOR_TYPE_RGB_ALPHA:      rgba = RGBAtoRGBA(pixels, width * height);              break;
            default:                            throw new Exception("loadPNG: Unknown pixel format.");
        }

        delete pixels;
        delete palette;

        return new Canvas(rgba, width, height);
    }
    finally
    {
        if (png !== null)
            png_destroy_read_struct(
                &png,
                (info !== null) ? &info : null,
                (end !== null) ? &end : null
            );
    }
}

void savePNG(Stream stream, Canvas canvas)
{
    png_struct* png_ptr = null;
    png_info* info_ptr = null;

    try
    {
        png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, null, null, null);
        if (png_ptr === null) throw new Exception("Failed to create write struct.");

        //setjmp(png_jmpbuf(png_ptr)); // hope this isn't important. :P

        info_ptr = png_create_info_struct(png_ptr);
        if (info_ptr === null) throw new Exception("Failed to create info struct.");

        int width = canvas.width();
        int height = canvas.height();

        png_set_write_fn(png_ptr, cast(void*)stream, &writePNG, &flushPNG);

        int colourFormat = PNG_COLOR_TYPE_RGB_ALPHA;
        int colourBpp = 4;

        png_set_IHDR(
            png_ptr, info_ptr,
            width, height,
            8,
            colourFormat,
            PNG_INTERLACE_NONE,
            PNG_COMPRESSION_TYPE_DEFAULT,
            PNG_FILTER_TYPE_DEFAULT);

        void** rows = cast(void**)png_malloc(png_ptr, (void*).size * height);
        for (int i = 0; i < height; i++)
        {
            rows[i] = png_malloc(png_ptr, colourBpp * width);
            memcpy(rows[i], canvas.getPointer(0, i), width * colourBpp);
        }
        png_set_rows(png_ptr, info_ptr, rows);
        info_ptr.valid |= PNG_INFO_IDAT;
        png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, null);
    }
    finally
    {
        if (png_ptr !== null)
            png_destroy_write_struct(&png_ptr, &info_ptr);
    }
}