/**
 * libPNG interface.
 *
 * Adapted version of Burton Raydon's image loader for dig.
 * Loads a PNG image into a software buffer, where it can be manipulated
 * freely.
 *
 * Tweaks by Andy Friesen.
 *
 * Some pluggable factory things could be done here so that more image
 * formats could be added more easily.
 *
 * I'd also like to eventually support more pixel formats than 32bpp ARGB.
 * I was thinking that I should store the pixel data as a void*, then have
 * a PixelFormatStrategy, which would be responsible for returning pixels
 * in some requested format, without ever forcing the user to manually
 * deal with pixel format conversions.
 *
 * Distribution / Code rights:
 *  Use this source code in any fashion you see fit.  Giving me credit where
 *  credit is due is optional, depending on your own levels of integrity and
 *  honesty.
 *
 */

import windows;
import stream;

/** A colour.  Currently stored in 32bpp ARGB format. */
struct Colour
{
align(1):
    ubyte b, g, r, a;

//--
    /** Named constructor idiom, since D doesn't allow structs to have
     *  constructors.
     * @param R Red intensity.
     * @param G Green intensity.
     * @param B Blue intensity.
     * @param A Alpha level. (255 is opaque, 0 is completely transparent)
     * @return The colour.
     */
    static Colour create(ubyte R, ubyte G, ubyte B, ubyte A)
    {
        Colour c;
        c.a = A;
        c.r = R;
        c.g = G;
        c.b = B;
        return c;
    }

    /** Same as above, but sets alpha to the maximum (the colour is opaque).
     * @param R Red intensity.
     * @param G Green intensity.
     * @param B Blue intensity.
     * @return The colour.
     */
    static Colour create(ubyte R, ubyte G, ubyte B)
    {
        return create(R, G, B, 255);
    }
}

/// Software representation of an image.
class Image
{
private:
    Colour* _pixels;
    int _width;
    int _height;

    this(Colour* p, int w, int h)
    {
        _pixels = p;
        _width = w;
        _height = h;
    }

public final:
    ~this()
    {
        delete _pixels; // maybe I should let the GC do this.
    }

    /** @return A pointer to the raw pixel data.  Do not delete this pointer! */
    Colour* pixels()
    {
        return _pixels;
    }

    /** @return The width of the image. */
    int width()
    {
        return _width;
    }

    /** @return The height of the image. */
    int height()
    {
        return _height;
    }

    /** @return Returns the pixel at (x, y) */
    Colour getPixel(int x, int y)
    {
        if (x < 0 || y < 0 || x >= _width || y >= _height)
            return Colour.create(0, 0, 0, 0);
        else
            return _pixels[y * _width + x];
    }

    /** @return The number of bytes required to store one pixel.
     * This is constant currently, but will later change if support for
     * other pixel formats is added.
     */
    int pixelSize()
    {
        return 4;
    }
}

private
{
    HANDLE library;
    const char[] libraryName = "libpng.dll";

    static this()
    {
        library = LoadLibraryA(libraryName);
        if (library == (HANDLE) 0)
            throw new Exception("Unable to load " ~ libraryName);

        load ((void **) &png_create_info_struct, "png_create_info_struct");
        load ((void **) &png_create_read_struct, "png_create_read_struct");
        load ((void **) &png_create_write_struct, "png_create_write_struct");
        load ((void **) &png_destroy_read_struct, "png_destroy_read_struct");
        load ((void **) &png_destroy_write_struct, "png_destroy_write_struct");
        load ((void **) &png_get_IHDR, "png_get_IHDR");
        load ((void **) &png_get_PLTE, "png_get_PLTE");
        load ((void **) &png_get_io_ptr, "png_get_io_ptr");
        load ((void **) &png_get_text, "png_get_text");
        load ((void **) &png_get_rowbytes, "png_get_rowbytes");
        load ((void **) &png_init_io, "png_init_io");
        load ((void **) &png_read_image, "png_read_image");
        load ((void **) &png_read_info, "png_read_info");
        load ((void **) &png_read_row, "png_read_row");
        load ((void **) &png_set_IHDR, "png_set_IHDR");
        load ((void **) &png_set_packing, "png_set_packing");
        load ((void **) &png_set_read_fn, "png_set_read_fn");
        load ((void **) &png_set_sig_bytes, "png_set_sig_bytes");
        load ((void **) &png_set_strip_16, "png_set_strip_16");
        load ((void **) &png_set_text, "png_set_text");
        load ((void **) &png_set_write_fn, "png_set_write_fn");
        load ((void **) &png_sig_cmp, "png_sig_cmp");
        load ((void **) &png_write_end, "png_write_end");
        load ((void **) &png_write_image, "png_write_image");
        load ((void **) &png_write_info, "png_write_info");
    }

    void load (void **pointer, char [] name)
    {
        *pointer = GetProcAddress (library, toStringz (name));
        if (*pointer === null)
            throw new Exception("couldn't find " ~ name ~ " in " ~ libraryName);
    }

    // Pixel format conversion routines.
    // length is the number of pixels, not bytes
    Colour[] i8ToRGBA(ubyte[] src, int length, Colour[] palette)
    {
        Colour[] dest = new Colour[length];
        for (int i = 0; i < length; i++)
            dest[i] = palette[src[i]];
        return dest;
    }

    Colour[] RGBtoRGBA(ubyte[] src, int length)
    {
        int j = 0;
        Colour[] dest = new Colour[length];
        for (int i = 0; i < length; i++)
        {
            ubyte b = 0;//src[j++];
            ubyte g = 0;//src[j++];
            ubyte r = 0;//src[j++];
            dest[i] = Colour.create(r, g, b);
        }
        return dest;
    }

    Colour[] grayScaleToRGBA(ubyte[] src, int length)
    {
        Colour[] dest = new Colour[length];
        for (int i = 0; i < length; i++)
            dest[i] = Colour.create(src[i], src[i], src[i]);
        return dest;
    }

    Colour[] grayScaleAlphaToRGBA(ubyte[] src, int length)
    {
        Colour[] dest = new Colour[length];
        int j = 0;
        for (int i = 0; i < length; i++)
        {
            ubyte l = src[j++];
            ubyte a = src[j++];
            dest[i] = Colour.create(l, l, l, a);
        }
        return dest;
    }

    // heh.  Maybe I should just cast or something.
    Colour[] RGBAtoRGBA(ubyte[] src, int length)
    {
        int j = 0;
        Colour[] dest = new Colour[length];
        for (int i = 0; i < length; i++)
        {
            ubyte b = src[j++];
            ubyte g = src[j++];
            ubyte r = src[j++];
            ubyte a = src[j++];
            dest[i] = Colour.create(r, g, b, a);
        }
        return dest;
    }
}

extern (C)
{
    struct png_struct
    {
    }

    struct png_info
    {
    }

    struct png_text
    {
        int compression; /* compression value:
                           -1: tEXt, none
                            0: zTXt, deflate
                            1: iTXt, none
                            2: iTXt, deflate  */
        char *key; /* keyword, 1-79 character description of "text" */
        char *text; /* comment, may be an empty string (ie "") */
        int text_length; /* length of the text string */
        int itxt_length; /* length of the itxt string */
        char *lang; /* language code, 1-79 characters */
        char *lang_key; /* keyword translated UTF-8 string, 0 or more chars */
    }

    struct png_color
    {
        ubyte red;
        ubyte green;
        ubyte blue;
    }

    typedef void (*png_error) (png_struct *, char *);
    typedef void (*png_flush) (png_struct *);
    typedef void (*png_rw) (png_struct *, byte *, int);

    const char *PNG_LIBPNG_VER_STRING = "1.0.6";

    /* These describe the color_type field in png_info. */
    /* color type masks */
    const int PNG_COLOR_MASK_PALETTE = 1;
    const int PNG_COLOR_MASK_COLOR = 2;
    const int PNG_COLOR_MASK_ALPHA = 4;

    /* color types.  Note that not all combinations are legal */
    const int PNG_COLOR_TYPE_GRAY = 0;
    const int PNG_COLOR_TYPE_PALETTE = (PNG_COLOR_MASK_COLOR | PNG_COLOR_MASK_PALETTE);
    const int PNG_COLOR_TYPE_RGB = (PNG_COLOR_MASK_COLOR);
    const int PNG_COLOR_TYPE_RGB_ALPHA = (PNG_COLOR_MASK_COLOR | PNG_COLOR_MASK_ALPHA);
    const int PNG_COLOR_TYPE_GRAY_ALPHA = (PNG_COLOR_MASK_ALPHA);

    /* These are for the interlacing type.  These values should NOT be changed. */
    const int PNG_INTERLACE_NONE = 0; /* Non-interlaced image */
    const int PNG_INTERLACE_ADAM7 = 1; /* Adam7 interlacing */
    const int PNG_INTERLACE_LAST = 2; /* Not a valid value */

    /* This is for compression type. PNG 1.0-1.2 only define the single type. */
    const int PNG_COMPRESSION_TYPE_BASE = 0; /* Deflate method 8, 32K window */
    const int PNG_COMPRESSION_TYPE_DEFAULT = PNG_COMPRESSION_TYPE_BASE;

    /* This is for filter type. PNG 1.0-1.2 only define the single type. */
    const int PNG_FILTER_TYPE_BASE = 0; /* Single row per-byte filtering */
    const int PNG_FILTER_TYPE_DEFAULT = PNG_FILTER_TYPE_BASE;

    /* Supported compression types for text in PNG files (tEXt, and zTXt).
     * The values of the PNG_TEXT_COMPRESSION_ defines should NOT be changed. */
    const int PNG_TEXT_COMPRESSION_NONE_WR = -3;
    const int PNG_TEXT_COMPRESSION_zTXt_WR = -2;
    const int PNG_TEXT_COMPRESSION_NONE = -1;
    const int PNG_TEXT_COMPRESSION_zTXt = 0;
    const int PNG_ITXT_COMPRESSION_NONE = 1;
    const int PNG_ITXT_COMPRESSION_zTXt = 2;
    const int PNG_TEXT_COMPRESSION_LAST = 3;  /* Not a valid value */

    png_info *(*png_create_info_struct) (png_struct *png);
    png_struct *(*png_create_read_struct) (char *user_png_ver, void *error_ptr, png_error *error_fn, png_error *warn_fn);
    png_struct *(*png_create_write_struct) (char *user_png_ver, void *error_ptr, png_error *error_fn, png_error *warn_fn);
    void (*png_destroy_read_struct) (png_struct **png, png_info **info, png_info **end);
    void (*png_destroy_write_struct) (png_struct **png, png_info **info);
    uint (*png_get_IHDR) (png_struct *png, png_info *info, uint *width, uint *height, int *bit_depth, int *color_type, int *interlace_type, int *compression_type, int *filter_type);
    uint (*png_get_PLTE) (png_struct *png, png_info *info, png_color **palette, int *num_palette);
    void *(*png_get_io_ptr) (png_struct *png);
    uint (*png_get_text) (png_struct *png, png_info *info, png_text **text, int *num_text);
    uint (*png_get_rowbytes) (png_struct* png, png_info* info);
    void (*png_init_io) (png_struct *png, FILE *fp);
    void (*png_read_image) (png_struct *png, byte **image);
    void (*png_read_info) (png_struct *png, png_info *info);
    void (*png_read_row) (png_struct *png, ubyte *row, ubyte *display_row);
    void (*png_set_IHDR) (png_struct *png, png_info *info, uint width, uint height, int bit_depth, int color_type, int interlace_type, int compression_type, int filter_type);
    void (*png_set_packing) (png_struct *png);
    void (*png_set_read_fn) (png_struct *png, void *io_ptr, png_rw read_data_fn);
    void (*png_set_sig_bytes) (png_struct *png, int num_bytes);
    void (*png_set_strip_16) (png_struct *png);
    void (*png_set_text) (png_struct *png, png_info *info, png_text *text, int num_text);
    void (*png_set_write_fn) (png_struct *png, void *io_ptr, png_rw write_data_fn, png_flush output_flush_fn);
    int (*png_sig_cmp) (ubyte *sig, int start, int num_to_check);
    void (*png_write_end) (png_struct *png, png_info *info);
    void (*png_write_image) (png_struct *png, byte **image);
    void (*png_write_info) (png_struct *png, png_info *info);

    extern (C)
    void read_data (png_struct *png, byte *data, int length)
    {
        Stream stream;

        stream = cast (Stream) png_get_io_ptr (png);
        stream.readExact (data, length);
    }

}

/** Loads an image.
 * @param fileName Filename of the image to load.
 * @return The loaded image.
 * @throw Exception if an error occurs.
 */
Image loadPNG(char[] fileName)
{
    Stream stream = new File(fileName, FileMode.In);
    Image img = loadPNG(stream);
    delete stream;

    return img;
}

/** Loads an image.
 * @param stream The stream from which to load the image.
 * @return The image.
 * @throw Exception if the image cannot be loaded.
 */
Image loadPNG(Stream stream)
{
    ubyte[8] header;
    if (stream.read(header) != 8)
        throw new Exception("Unable to read PNG header.");

    if (png_sig_cmp(header, 0, header.length) != 0)
        throw new Exception("File is not a PNG image.");

    png_struct* png = null;
    png_info* info = null;
    png_info* end = null;

    try
    {
        // Initial libpng bootup.
        png = png_create_read_struct (PNG_LIBPNG_VER_STRING, null, null, null);
        if (png === null)
            throw new Exception("Couldn't allocate png struct.");

        info = png_create_info_struct (png);
        if (info === null)
            throw new Exception("Couldn't allocate PNG info block");

        end = png_create_info_struct (png);
        if (end === null)
            throw new Exception("Couldn't allocate PNG end info block");

        png_set_read_fn(png, cast(void*) stream, &read_data);
        //png_init_io (png, file);
        png_set_sig_bytes(png, header.length);

        uint width, height;
        int depth;
        int colorType;
        int filterType;
        int compressionType;
        int interlaceType;

        png_read_info(png, info);
        png_get_IHDR(png, info, &width, &height, &depth, &colorType,
            &interlaceType, &compressionType, &filterType);

        int pitch = png_get_rowbytes(png, info);
        ubyte[] pixels = new ubyte[width * height * depth];
        Colour[] palette = null;

        // extract and convert palette
        if (colorType == PNG_COLOR_TYPE_PALETTE)
        {
            png_color* colours;
            int count;
            png_get_PLTE(png, info, &colours, &count);
            palette = new Colour[count];

            for (int c = 0; c < count; c++)
            {
                png_color* s = &colours[c];
                palette[c] = Colour.create(s.red, s.blue, s.green); // FIXME: depending on size of pixels here.
            }
        }

        // read the actual pixels
        {
            ubyte* row = &pixels[0];
            for (int y = 0; y < height; y++, row += pitch)
                png_read_row(png, row, null);
        }

        Colour[] rgba;
        switch (colorType)
        {
            case PNG_COLOR_TYPE_GRAY:           rgba = grayScaleToRGBA(pixels, width * height);         break;
            case PNG_COLOR_TYPE_GRAY_ALPHA:     rgba = grayScaleAlphaToRGBA(pixels, width * height);    break;
            case PNG_COLOR_TYPE_PALETTE:        rgba = i8ToRGBA(pixels, width * height, palette);       break;
            case PNG_COLOR_TYPE_RGB:            rgba = RGBtoRGBA(pixels, width * height);               break;
            case PNG_COLOR_TYPE_RGB_ALPHA:      rgba = RGBAtoRGBA(pixels, width * height);              break;
        }

        delete pixels;
        delete palette;

        return new Image(rgba, width, height);
    }
    finally
    {
        if (png !== null)
            png_destroy_read_struct(
                &png,
                (info !== null) ? &info : null,
                (end !== null) ? &end : null
            );
    }
}